#include <pybind11/pybind11.h>
#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

#define BLOCK_SIZE 256
#define WARP_SIZE 32

// This kernel fuses ReLU and GroupNorm with minimized warp divergence.
// It refactors conditional logic into uniform loops and uses vectorized memory accesses for the main body.

__global__ void fused_relu_groupnorm_no_div_kernel(
    float* __restrict__ data,            // tensor of shape [N, C, D, H, W]
    const float* __restrict__ gamma,       // GroupNorm weight, shape: [C]
    const float* __restrict__ beta,        // GroupNorm bias, shape: [C]
    int N, int C, int D, int H, int W,
    int G, float eps)                      // groups and epsilon
{
    // Compute tensor dimensions
    int spatial_size = D * H * W;
    int channels_per_group = C / G;

    // Identify current sample and group
    int n = blockIdx.x;  // sample index
    int g = blockIdx.y;  // group index
    int c_start = g * channels_per_group;

    // Pointer to the current group's data
    int group_offset = n * (C * spatial_size) + c_start * spatial_size;
    float* group_data = data + group_offset;
    int group_elems = channels_per_group * spatial_size;

    // Allocate shared memory:
    // - First: load gamma and beta for this group (each array of size channels_per_group)
    // - Next: two arrays for warp reduction (WARP_SIZE each for sum and sumsq)
    extern __shared__ float s_mem[];
    float* sh_gamma = s_mem;                         // size: channels_per_group floats
    float* sh_beta  = sh_gamma + channels_per_group;   // size: channels_per_group floats
    float* s_sum    = sh_beta + channels_per_group;    // size: WARP_SIZE floats
    float* s_sumsq  = s_sum + WARP_SIZE;               // size: WARP_SIZE floats

    // Uniformly load GroupNorm parameters into shared memory
    for (int i = threadIdx.x; i < channels_per_group; i += blockDim.x) {
        sh_gamma[i] = gamma[g * channels_per_group + i];
        sh_beta[i]  = beta[g * channels_per_group + i];
    }
    __syncthreads();

    // First pass: Apply ReLU and accumulate sum and sum-of-squares.
    float local_sum = 0.f;
    float local_sumsq = 0.f;

    // Process as many elements as possible with vectorized loads (float4) for coalesced access
    int num_vec = group_elems >> 2;  // equivalent to group_elems / 4
    float4* group_data4 = reinterpret_cast<float4*>(group_data);

    for (int i = threadIdx.x; i < num_vec; i += blockDim.x) {
        float4 v = group_data4[i];
        // Apply ReLU in a branchless manner
        v.x = fmaxf(v.x, 0.f);
        v.y = fmaxf(v.y, 0.f);
        v.z = fmaxf(v.z, 0.f);
        v.w = fmaxf(v.w, 0.f);
        group_data4[i] = v;
        
        float sum_val = v.x + v.y + v.z + v.w;
        float sumsq_val = v.x*v.x + v.y*v.y + v.z*v.z + v.w*v.w;
        local_sum += sum_val;
        local_sumsq += sumsq_val;
    }

    // Process remaining elements (if group_elems is not divisible by 4)
    int start = num_vec << 2; // start index for remainder
    int rem = group_elems - start;  // remainder count (<4)
    int num_rem_iters = (rem + blockDim.x - 1) / blockDim.x;
    for (int i = 0; i < num_rem_iters; i++) {
        int idx = start + i * blockDim.x + threadIdx.x;
        if (idx < group_elems) {
            float val = group_data[idx];
            float r = fmaxf(val, 0.f);
            group_data[idx] = r;
            local_sum += r;
            local_sumsq += r * r;
        }
    }

    // Warp-level reduction using shuffle intrinsics for uniform control flow
    unsigned int mask = 0xffffffff;
    int lane = threadIdx.x & (WARP_SIZE - 1);
    for (int offset = WARP_SIZE / 2; offset > 0; offset /= 2) {
        local_sum  += __shfl_down_sync(mask, local_sum, offset);
        local_sumsq += __shfl_down_sync(mask, local_sumsq, offset);
    }

    // Write warp-level results to shared memory
    int warp_id = threadIdx.x / WARP_SIZE;
    if (lane == 0) {
        s_sum[warp_id] = local_sum;
        s_sumsq[warp_id] = local_sumsq;
    }
    __syncthreads();

    // Final block-level reduction by thread 0
    if (threadIdx.x == 0) {
        float group_sum = 0.f;
        float group_sumsq = 0.f;
        int num_warps = (blockDim.x + WARP_SIZE - 1) / WARP_SIZE;
        for (int i = 0; i < num_warps; i++) {
            group_sum  += s_sum[i];
            group_sumsq += s_sumsq[i];
        }
        float mean = group_sum / group_elems;
        float var = group_sumsq / group_elems - mean * mean;
        float inv_std = rsqrtf(var + eps);
        s_sum[0] = mean;       // broadcast mean
        s_sumsq[0] = inv_std;  // broadcast inverse std
    }
    __syncthreads();

    float mean = s_sum[0];
    float inv_std = s_sumsq[0];

    // Second pass: Normalize the data using loaded gamma and beta from shared memory
    for (int i = threadIdx.x; i < group_elems; i += blockDim.x) {
        int channel = i / spatial_size;  // Determine channel index within the group
        float g_val = sh_gamma[channel];
        float b_val = sh_beta[channel];
        float norm_val = (group_data[i] - mean) * inv_std;
        group_data[i] = norm_val * g_val + b_val;
    }
}

// Forward function: First performs 3D transposed convolution, then applies the fused kernel

torch::Tensor forward(
    torch::Tensor x,
    torch::Tensor conv_transpose,
    torch::Tensor group_norm_weight,
    torch::Tensor group_norm_bias,
    int64_t groups,
    double eps) {

    // Perform 3D transposed convolution using the ATen operator
    auto y = at::conv_transpose3d(
        x,
        conv_transpose,
        /*bias=*/c10::nullopt,
        /*stride=*/{1, 1, 1},
        /*padding=*/{0, 0, 0},
        /*output_padding=*/{0, 0, 0},
        /*groups=*/1,
        /*dilation=*/{1, 1, 1}
    );

    int N = y.size(0);
    int C = y.size(1);
    int D = y.size(2);
    int H = y.size(3);
    int W = y.size(4);
    int G = groups;
    int channels_per_group = C / G;
    int spatial_size = D * H * W;

    // Calculate shared memory size in bytes:
    // - Two arrays for gamma and beta: 2 * channels_per_group floats
    // - Two warp reduction arrays: 2 * WARP_SIZE floats
    size_t shared_mem_size = (channels_per_group * 2 + 2 * WARP_SIZE) * sizeof(float);

    dim3 grid(N, G);
    dim3 block(BLOCK_SIZE);

    fused_relu_groupnorm_no_div_kernel<<<grid, block, shared_mem_size>>>(
         y.data_ptr<float>(),
         group_norm_weight.data_ptr<float>(),
         group_norm_bias.data_ptr<float>(),
         N, C, D, H, W,
         G, static_cast<float>(eps)
    );

    cudaDeviceSynchronize();
    return y;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Fused ConvTranspose3D + ReLU + GroupNorm with minimized warp divergence (CUDA)");
}
